#!/usr/bin/perl

################################################################################
# rsync_to_shadow.pl
#
# This script rsyncs a master radmind server to a slave
#
# Copyright (c) 2002 University of Utah Student Computing Labs.
# All Rights Reserved.
#
# Permission to use, copy, modify, and distribute this software and
# its documentation for any purpose and without fee is hereby granted,
# provided that the above copyright notice appears in all copies and
# that both that copyright notice and this permission notice appear
# in supporting documentation, and that the name of The University
# of Utah not be used in advertising or publicity pertaining to
# distribution of the software without specific, written prior
# permission. This software is supplied as is without expressed or
# implied warranties of any kind.
#
# 2008.10.31 - Added checking so that only one copy of this script runs.
# 2006.12.21 - No longer deletes old overload catalog.  So it will keep the old
#              version until a successful sync, and it will back it up anyway.
# 2006.11.02 - Rsync folders in /var/radmind/transcript/ better.  Update
#              catalog only if rsync is successful.
# 2005.12.30 - Now keeps a catalog of overloads and their checksum. This catalog
#              is synced to the other server.  So the other server can read the
#              file and delete overloads not listed in it!
#              This script will also only sync overloads with changed
#              transcripts to make it much faster (hopefully!)
# 2005.??.?? - Changed it so it treats /var/radmind/transcript the same as 
#              /var/radmind/file
# 2005.01.05 - Changed how /var/radmind/file was handled.  Now I can tell this 
#              script to scan the entire file directory or only each folder in
#              the file directory.
#
################################################################################

################################################################################
# Instructions
# For this script to be able to run un-attended (via crontab), the server this
# machine is connecting to must have ssh setup so that I can shh into it without
# entering a password.  Yes, a security risk, but if you take precautions (use
# a private network, lock down ssh access via tcp-wrappers/firewall/sshd_config
# Allow attribute, authorized_keys command wrapper, non-root user then you should
# be safe).
#
# On the master server, run ssh-keygen (for ssh2).  When it asks for the password
# press return.  Do not enter a password or else you can't automate it!  It will
# create keys in "$HOME/.ssh". Copy the file "$HOME/.ssh/id_rsa.pub" to the
# shadow server.  Protect "$HOME/.ssh/id_rsa" as anyone who has it can imiate your
# master server now.
#
# On the shadow server, you need to have a user folder that this script will login
# as.  I just create a user named "radmind" in System Prefs.  Move the id_rsa.pub
# file copied from the master server to /Users/radmind/.ssh/ and name it
# authorized_keys.
#
# On the master server, ssh into the shadow server as the radmind user.  It
# shouldn't ask for a password.  If it does, repeat the steps until it doesn't.
#
# Make sure the radmind user can read all of the content in /var/radmind/commmand
# /var/radmind/file, /var/radmind/transcript, and /var/radmind/config.  If you run
# this script as root, then you have that problem solved.
#
# Lastly, on the shadow server, you can add additional security by using a command
# wrapper.  Do that by using the script rsync-wrapper.sh (which hopefully you have).
# Add this text on the shadow server to the file /Users/radmind/.ssh/authorized_keys
#
# from="10.0.1.1",command="/etc/rsync-wrapper.sh" ssh-ras ....(key)....
#
# The ....(key).... is replaced the public key from the file id_rsa.pub that was
# copied from the master machine.  They have to be one line.
#
# Naturally, if you use a private network (either ethernet port to ethernet port
# or the slower firewire networking) you can use 10.0.1.1 IP (which is what we do).
#
################################################################################

use Fcntl qw/O_WRONLY O_CREAT O_EXCL/;

my $debug = 1;

################################################################################
# Change these
#
$source_base = "/var/radmind";
$destination_base = "radmind\@10.0.1.2:/var/radmind";
$logger = "/usr/bin/logger -i -t $0"; # Your logger
$noisy_log = 0;
$rsync_output = "/var/log/rsync_output"; # this should be a folder
if ( ! -d $rsync_output ) {
  print "The directory $rsync_output is missing.  Creating it.\n" if $debug;
  if ( -e $rsync_output ) {
    unlink $rsync_output;
  }
  mkdir $rsync_output;
}

$rsync_output_backup_number = 5;

@rsync_these_folders = ( "command", "config", "special", "overload_catalog" ); 
# Do not include "file" or "transcript" in the above list, because it is handled with 
# the $scan_file_folder_seperately variable.

$scan_file_folder_seperately = 1;
if ( defined $ARGV[0] ) {
  print "Scanning everything in one huge attempt.\n" if $debug;
  $scan_file_folder_seperately = 0;
} else {
  print "Scanning one overload at a time.\n" if $debug;
}

################################################################################
# Internal stuff
#

$rsync_command = "/usr/bin/rsync -vrl --rsh=ssh --hard-links --progress --delete ";   
$rsync_command_non_recursive = "/usr/bin/rsync -vrl --rsh=ssh --hard-links --progress --delete ";   

$cl_to_gui_messenger = "/Applications/CL_to_GUI_Messenger.app/Contents/MacOS/CL_to_GUI_Messenger";

$last_source = "bla"; # don't change this
$output_file = "$rsync_output/rsync.out";
$error_file = "$rsync_output/rsync.err";
$error_bool = 0;
$last_error_message = "";

################################################################################
# Start
#

system "$logger \"Started.\""; 

################################################################################
# Quit if another one of me is running.
#

my $run_file = "/var/run/rsync_to_shadow";
my $wait = 1;
my $counter = 1;
while ( $wait and $counter < 10 ) {
  $wait = 0;
  $counter++;
  if ( open(FH, "<", $run_file) ) {
    print "I'm already running... ($run_file exists) Waiting 60 seconds.\n" if $debug;
    system "$logger \"I'm already running... Waiting 60 seconds.\"";
    $wait = 1;
    close(FH);
    sleep 60;
  } else {
    sysopen(FH, $run_file, O_WRONLY | O_CREAT | O_EXCL) or $wait = 2;
  }
  if ( $wait == 2 ) {
    system "$logger \"Can't create file $run_file. Waiting 60 seconds.\"";
    print "Can't create file $fun_file. Waiting 60 seconds.\n" if $debug;
    sleep 60;
  }
}
if ( $wait ) {
  system "$logger \"Quiting because another one of me is running.\"";
  print "Quiting because another one of me is running.\n" if $debug;
  sleep 1;
  exit 1;
}

################################################################################
# Log
#

if ( ! -d $rsync_output ) {
  system "$logger \"No valid log location, exiting.\""; 
  print "No valid log location, exiting.\n" if $debug;
  exit 1;
}

# Handle the log file
roll_log ( "$output_file", "$rsync_output_backup_number" );
roll_log ( "$error_file", "$rsync_output_backup_number" );

################################################################################
# Get list of folders to sync
#

if ( $scan_file_folder_seperately ) {

  my $catalog_file_backup = "$source_base/backup/overload_catalog";
  my $catalog_file_old = "$source_base/overload_catalog";
  my $catalog_file_new = "$source_base/overload_catalog_new";
  my $folder_to_catalog = "$source_base/transcript";
  unlink "$source_base/transcript/.DS_Store";

  # Read old overload catalog
  my %old_overload_catalog = ();
  if ( -e $catalog_file_old ) {
    open ( FILE, "< $catalog_file_old" ) or die "Can't read from $catalog_file_old : $!";
    while( <FILE> ) {
      my ( $o_path, $o_cksum ) = ( split /\s+/, $_ );
      $old_overload_catalog{ $o_path } = $o_cksum;
    }
    close FILE;
  }

  print "Scanning overloads\n";

  # scan all overloads
  @transcripts = getFileFolderList ( $folder_to_catalog );
  my @new_overload_catalog = ();
  foreach $transcript ( @transcripts ) {
    
    my $fsdiff = `/usr/local/bin/fsdiff -c sha1 -1 \"$folder_to_catalog/$transcript\"`;
    my ( $n_type, $n_path, $n_perm, $n_owner, $n_group, $n_mtime, $n_size, $n_cksum ) = ( split /\s+/, $fsdiff );

    # rsync list of new or updated overloads
    if ( not ( defined $old_overload_catalog {$n_path} and $n_cksum eq $old_overload_catalog {$n_path} ) ) {

      # sync parent folders
        @parentList = split '/', $transcript;
        $parentsString = "";
        for ( $ii = 0; $ii < $transcript =~ tr/\///; $ii++ ) {
          $parentsString .= $parentList[$ii]."/";
          print $parentsString."\n";
          run_rsync ( $rsync_command_non_recursive, "$source_base/transcript/$parentsString", "$destination_base/transcript/$parentsString" );
        if ( -d "$source_base/file/$transcript" ) {
          run_rsync ( $rsync_command_non_recursive, "$source_base/file/$parentsString", "$destination_base/file/$parentsString" );
        }
        }
      
      # rsync the real deal (transcripts)
      my $unixResult1 = run_rsync ( $rsync_command, "$source_base/transcript/$transcript", "$destination_base/transcript/$transcript" );
      my $unixResult2;
      if ( -d "$source_base/file/$transcript" ) {
        $unixResult2 = run_rsync ( $rsync_command, "$source_base/file/$transcript", "$destination_base/file/$transcript" );
      }

      # check for success      
      if ( $unixResult1 == 0 and $unixResult2 == 0 ) { # 0 means no error for "unix command results"
        # record successful overload sync to catalog file
        open ( FILE, ">> $catalog_file_new" ) or die "Can't write to $catalog_file_new : $!";
        print FILE $n_path ." ". $n_cksum ."\n";
        close FILE;
      }
    } else {
      # record previously synced overload to catalog file
      open ( FILE, ">> $catalog_file_new" ) or die "Can't write to catalog_file_new : $!";
      print FILE $n_path ." ". $n_cksum ."\n";
      close FILE;

    }
  }
  
  system "/bin/mkdir", "-r", $catalog_file_backup if ! -e $catalog_file_backup;
  system "/bin/mv", $catalog_file_old, $catalog_file_backup;
  system "/bin/mv", $catalog_file_new, $catalog_file_old;
  
} else {
  run_rsync ( $rsync_command, "$source_base/file", "$destination_base/file" );
  run_rsync ( $rsync_command, "$source_base/transcript", "$destination_base/transcript" );
}

################################################################################
# Run rsync on the rest
#
foreach $i ( @rsync_these_folders ) {
  run_rsync ( $rsync_command, "$source_base/$i", "$destination_base/$i" );
}

################################################################################
# Clean up
#

# Erase last message
if ( -e $cl_to_gui_messenger ) {
  if ( $last_source ne "bla" ) {
    killCLGUI ( "$cl_to_gui_messenger", "root", "$last_source" );
  }
}

# Display message if there was an error
if ( -e $cl_to_gui_messenger ) {
  if ( $error_bool ) {
    system "\"$cl_to_gui_messenger\" \"There was an rsync error.\" \"$last_error_message\" &";
  }
}

system "$logger \"Finished.\"";
unlink $run_file;


################################################################################
sub run_rsync {
  my ( $rsync_command, $source, $destination ) = @_;

  # Do GUI message
  if ( -e $cl_to_gui_messenger ) {
    if ( $last_source ne "bla" ) {
      killCLGUI ( "$cl_to_gui_messenger", "root", "$last_source" );
    }
    $last_source = $source;
    system "\"$cl_to_gui_messenger\" \"Running Rsync, please do not change any radmind stuff!\" \"$source\" &";
  }

  # check the to see if it is a directory and append "/" if it is...
  if ( -d $source ) {
    $source .= "/";
    $destination .= "/";
    $source =~ s'//+'/';
    $destination =~ s'//+'/';
  }

  # Print log info
  open ( LOG, ">>$output_file" );
  print LOG `date`;
  print LOG "$source_base/$i\n";
  print LOG "$rsync_command \"$source\" \"$destination\"\n\n";
  #print "$rsync_command \"$source\" \"$destination\"\n\n";

  # Run Rsync
  if ( $noisy_log ) {
    open ( COMMAND, "$rsync_command \"$source\" \"$destination\" 2>> $error_file |" );
    while ( <COMMAND> ) {
      print LOG;
    }
    close ( COMMAND );
  } else {
    system "$rsync_command \"$source\" \"$destination\" 2>> $error_file";  
  }

  # Check for errors
  $result = $? >> 8;
  if ( $result > 0 ) {
    $error_bool = 1;
    system "$logger \"Error: $source\"";
    $last_error_message .= $source;
    print LOG "There was an error.";
  }
  close ( LOG );
  return $result;
}

################################################################################
sub getFileFolderList {
  my ( $folder ) = @_;
  my @dirlist=();
  if ( -d $folder ) {
    @dirlist=findFiles($folder);
    foreach $i ( @dirlist ) {
      $i =~ s/$folder\///;
    }
  } 
  return @dirlist;
}

################################################################################
sub findFiles {
  my $dir = shift;
  my @tempy = ();
  opendir DIR, $dir or return;
  my @contents =
        map "$dir/$_",
        sort grep !/^\.\.?$/,
        readdir DIR;
  closedir DIR;
  foreach (@contents) {
        if ( !-l && -d ) {
          push @tempy, &findFiles($_);
        } else {
      push @tempy, $_;
        }
  }
  return @tempy;
}

################################################################################
# This script renames $path_to_log so that the lowest number is always the
# newest.  The oldest that is greater than $number_of_backups is deleted.
#  
sub roll_log {
  my ( $path_to_log, $number_of_backups ) = @_;
  if ( -e $path_to_log ) {
    if ( -e $path_to_log.".bak$number_of_backups" ) {    
      unlink $path_to_log.".bak$number_of_backups";
    }
    for ( $i = $number_of_backups ; $i > 1 ; $i-- ) {
      if ( -f $path_to_log.".bak".($i-1)) {
        system "/bin/mv -f \"$path_to_log".".bak".($i-1)."\" \"$path_to_log".".bak$i\"";
      }
    }
    system "/bin/mv -f \"$path_to_log\" \"$path_to_log".".bak1\"\n";
  } 
}

################################################################################
sub killCLGUI  {
  my ( $appToCheck, $username, $message_to_grep_for ) = @_;
  open ( PROC, "ps -wwxo pid,command -U $username |" );   
  @proclist = <PROC>;
  close PROC;
  @cl_to_gui_list = grep ( /$appToCheck/i, @proclist );
  @kill_me = grep ( /$message_to_grep_for/i, @cl_to_gui_list );
  @pid = split ( ' ', $kill_me[0] );
  if ( $pid[0] ne "" ) {
    system "kill $pid[0]";
  }
}
