/*
gcc -arch i386 -arch ppc -O3 -o SetDisplay SetDisplay.c -framework Cocoa



SetDisplay.c

This tool sets the display resolution.  This tool exists so that we
can set the display to a setting when the display manager says it can't,
mainly, when a Mac starts up and the monitor is off, or if the Mac is on
a KVM.

Copyright (c) 2007 University of Utah Student Computing Labs.
All Rights Reserved.

Permission to use, copy, modify, and distribute this software and
its documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appears in all copies and
that both that copyright notice and this permission notice appear
in supporting documentation, and that the name of The University
of Utah not be used in advertising or publicity pertaining to
distribution of the software without specific, written prior
permission. This software is supplied as is without expressed or
implied warranties of any kind.

USAGE:
You can give it arguments or not.  If you don't give it arguments, it will use:

1024 x 768, 32 bits per pixel, and 75 htz

If you give it 4 arguments, they should be: width, height, bits per pixel, and refresh rate.

Like so:

SetDisplay 1024 768 32 75

WARNING:
In testing garbage values, I did get this to tool to change the display so that absolutely
nothing displayed.  I don't remember what I did to get that.  And I can't duplicate it anymore.
I could remotely control it, so I could fix it.  But beware, if you can't remotely control the
machine and fix bad display settings, well, you might have to resort to drastic measures to fix
a mac with bad display settings (like boot off of a different disk and delete the display 
preferences.

In other words, this tool will change the display settings regardless of what the display
manager says is possible.  So be careful!!!

*/

#include <ApplicationServices/ApplicationServices.h>
#include <CoreFoundation/CoreFoundation.h>
#include <IOKit/IOKitLib.h>
#include <stdlib.h>
#include <unistd.h>

#define MAX_DISPLAYS 32

typedef struct
{
	size_t width;
	size_t height;
	size_t bitsPerPixel;
	CGRefreshRate refresh;
} displayMode;

displayMode myMode;

static int numberForKey( CFDictionaryRef desc, CFStringRef key )
{
	CFNumberRef value;
	int num = 0;

	if ( (value = CFDictionaryGetValue(desc, key)) == NULL )
		return 0;
	CFNumberGetValue(value, kCFNumberIntType, &num);
	return num;
}

static void printDispDesc( CFDictionaryRef desc )
{
	char * msg;
	if ( CFDictionaryGetValue(desc, kCGDisplayModeUsableForDesktopGUI) == kCFBooleanTrue )
		msg = "Supports Aqua GUI";
	else
		msg = "Not for Aqua GUI";

	printf( "\t%d x %d,\t%d BPP,\t%d Hz\t(%s)\n",
			numberForKey(desc, kCGDisplayWidth),
			numberForKey(desc, kCGDisplayHeight),
			numberForKey(desc, kCGDisplayBitsPerPixel),
			numberForKey(desc, kCGDisplayRefreshRate),
			msg );
}

static void printShortDispDesc( CFDictionaryRef desc, int showOnlyAqua )
{
	int aqua;
	if ( CFDictionaryGetValue(desc, kCGDisplayModeUsableForDesktopGUI) == kCFBooleanTrue )
		aqua = 1;
	else
		aqua = 0;

	if ( showOnlyAqua == 1 ) {
		if ( aqua == 1 )
			printf( "%d %d %d %d Usable\n",
					numberForKey(desc, kCGDisplayWidth),
					numberForKey(desc, kCGDisplayHeight),
					numberForKey(desc, kCGDisplayBitsPerPixel),
					numberForKey(desc, kCGDisplayRefreshRate)
					);
	} else {
		printf( "%d %d %d %d %s\n",
				numberForKey(desc, kCGDisplayWidth),
				numberForKey(desc, kCGDisplayHeight),
				numberForKey(desc, kCGDisplayBitsPerPixel),
				numberForKey(desc, kCGDisplayRefreshRate),
				(aqua) ? "Usable" : "Nonusable"
				);	
	}
}

static void findClosestModeForDisplay( CGDirectDisplayID display, int verbose)
{
	CFDictionaryRef mode;
	boolean_t exactMatch;

	if ( verbose == 1 )
		printf( "Looking for chosest match to %ld x %ld, %ld Bits Per Pixel, %ld Hz\n", myMode.width, myMode.height, myMode.bitsPerPixel, (int)myMode.refresh );
	mode = CGDisplayBestModeForParametersAndRefreshRate(display, myMode.bitsPerPixel, myMode.width, myMode.height, myMode.refresh, &exactMatch);
	if ( verbose == 1 )
		printf ("The chosen mode:\n");
	
	if ( verbose == 1 )
		printDispDesc( mode );
	else
		printShortDispDesc( mode, 0 );
	
	if( !exactMatch && verbose == 1 )
	{
		printf( "Not an exact match\n" );
	}
}

static void findExactModeForDisplay( CGDirectDisplayID display, int verbose)
{
	CFIndex index, count;
	CFArrayRef modes;
	CFDictionaryRef mode;

	modes = CGDisplayAvailableModes (display); 
	count = CFArrayGetCount (modes);
	for (index = 0; index < count; index++) 
	{
		mode = CFArrayGetValueAtIndex( modes, index );
		if ( numberForKey(mode, kCGDisplayWidth) == myMode.width &&
				numberForKey(mode, kCGDisplayHeight) == myMode.height &&
				numberForKey(mode, kCGDisplayRefreshRate) == myMode.refresh &&
				numberForKey(mode, kCGDisplayBitsPerPixel) == myMode.bitsPerPixel)
		{
			if ( verbose == 1 )
				printDispDesc( mode );
			else
				printShortDispDesc( mode, 0 );
			return;
		}
	}
}

static void allModesForDisplay( CGDirectDisplayID display, int verbose)
{
	CFIndex index, count;
	CFArrayRef modes;
	CFDictionaryRef mode;
	printf( "------ All modes for display ------\n" );
	modes = CGDisplayAvailableModes (display); 
	count = CFArrayGetCount (modes);
	for (index = 0; index < count; index++) 
	{
		mode = CFArrayGetValueAtIndex( modes, index );
		printShortDispDesc( mode, verbose );
	}
	printf( "------------\n" );
}

static void highestModeForDisplay( CGDirectDisplayID display, displayMode *highestMode, int verbose )
{
	CFIndex index, count;
	CFArrayRef modes;
	CFDictionaryRef mode;
	
	highestMode->width = 0;
	highestMode->height = 0;
	highestMode->bitsPerPixel = 0;
	highestMode->refresh = 0;

	modes = CGDisplayAvailableModes (display); 
	count = CFArrayGetCount (modes);
	for (index = 0; index < count; index++) 
	{
		mode = CFArrayGetValueAtIndex( modes, index );
		if ( numberForKey(mode, kCGDisplayWidth) > highestMode->width )
			highestMode->width = numberForKey(mode, kCGDisplayWidth);
		if ( numberForKey(mode, kCGDisplayHeight) > highestMode->height )
			highestMode->height = numberForKey(mode, kCGDisplayHeight);
		if ( numberForKey(mode, kCGDisplayRefreshRate) > highestMode->refresh )
			highestMode->refresh = numberForKey(mode, kCGDisplayRefreshRate);
		if ( numberForKey(mode, kCGDisplayBitsPerPixel) > highestMode->bitsPerPixel )
			highestMode->bitsPerPixel = numberForKey(mode, kCGDisplayBitsPerPixel);
	}
	if ( verbose == 1 )
		printf( "Highest mode for display: %ld x %ld, %ld Bits Per Pixel, %ld Hz\n", highestMode->width, highestMode->height, highestMode->bitsPerPixel, (int)highestMode->refresh );
	else
		printf( "%ld %ld %ld %ld\n", highestMode->width, highestMode->height, highestMode->bitsPerPixel, (int)highestMode->refresh );
}

static void setdisplay( CGDirectDisplayID display, int verbose )
{
	CGError err;
	CGDisplayConfigRef configRef;
	CFDictionaryRef mode;
	boolean_t exactMatch;

	CGBeginDisplayConfiguration(&configRef);
	mode = CGDisplayBestModeForParametersAndRefreshRate(display, myMode.bitsPerPixel, myMode.width, myMode.height, myMode.refresh, &exactMatch);
	err = CGConfigureDisplayMode(configRef, display, mode);
	CGCompleteDisplayConfiguration( configRef, kCGConfigurePermanently );
	if ( err != CGDisplayNoErr )
	{
		printf( "Oops!  Mode switch failed?!?? (%d)\n", err );
	}
}

static void usage()
{
	printf( "SetDisplay [-acvxyz] [-w WIDTH] [-h HEIGHT] [-b BPP] [-r REFRESH] | [WIDTH HEIGHT BPP REFRESH]\n" );
	printf( " -a Show all possible matches (resolution not changed)\n" );
	printf( " -c Show closest match (resolution not changed)\n" );
	printf( " -v Verbose\n" );
	printf( " -x Show exact match (resolution not changed)\n" );
	printf( " -y Show and set highest possible resolution\n" );
	printf( " -z Show highest possible resolution (resolution not changed)\n" );
	printf( " No args default to 1024 768 32 75\n" );
	printf( " Values set with flags override values not set with flags (those last values are for legacy)\n" );
	exit(1);
}

int main(int argc, char **argv)
{
	CGDirectDisplayID displays[MAX_DISPLAYS];
	CGDisplayCount numDisplays;
	CGDisplayCount ii;
	CGDisplayErr err;
	int c;
	int verbose, shouldFindHighest, shouldFindExact, shouldShowAll, shouldFindClosest = 0;
	int shouldSetDisplay = 1;


	opterr = 0;

	myMode.width = 1024;
	myMode.height = 768;
	myMode.bitsPerPixel = 32;
	myMode.refresh = 75;

	while ((c = getopt (argc, argv, "ab:ch:r:vw:xyz")) != -1)
		printf ("Options %c\n", c);
		switch (c)
			{
			case 'w':
				myMode.width = atoi(optarg);
				break;
			case 'h':
				myMode.height = atoi(optarg);
				break;
			case 'b':
				myMode.bitsPerPixel = atoi(optarg);
				break;
			case 'r':
				myMode.refresh = atoi(optarg);
				break;
			case 'a':
				shouldShowAll = 1;
				shouldSetDisplay = 0;
				break;
			case 'c':
				shouldFindClosest = 1;
				shouldSetDisplay = 0;
				break;
			case 'v':
				verbose = 1;
				break;
			case 'x':
				shouldFindExact = 1;
				shouldSetDisplay = 0;
				break;
			case 'y':
				shouldFindHighest = 1;
				break;
			case 'z':
				shouldFindHighest = 1;
				shouldSetDisplay = 0;
				break;
			case '?':
				usage();
			//default:
				//abort ();
			}

	
	if ( argc > optind )
	{
		if ( argc == 4+optind )
		{
			myMode.width = atoi(argv[0+optind]);
			myMode.height = atoi(argv[1+optind]);
			myMode.bitsPerPixel = atoi(argv[2+optind]);
			myMode.refresh = atoi(argv[3+optind]);
		} else {
			usage();
		}
	}

	if ( verbose == 1 )
		printf( "Width: %d Height: %d BitsPerPixel: %d Refresh rate: %d\n", (int)myMode.width, (int)myMode.height, (int)myMode.bitsPerPixel, (int)myMode.refresh );

	//err = CGGetActiveDisplayList(MAX_DISPLAYS,
	err = CGGetOnlineDisplayList(MAX_DISPLAYS,
 								displays,
 								&numDisplays);
	if ( err != CGDisplayNoErr )
	{
		printf("Cannot get displays (%d)\n", err);
		exit( 1 );
	}

	if ( verbose == 1 )
		printf( "%d online display(s) found\n", (int)numDisplays );
		
	for (ii = 0; ii < numDisplays; ii++) 
	{
		CFDictionaryRef originalMode;
		if ( verbose == 1 && ! shouldShowAll )
			printf( "------------------------------------\n");
		originalMode = CGDisplayCurrentMode( displays[ii] );
		if ( originalMode == NULL )
		{
			printf( "Display 0x%x is invalid\n", (unsigned int)displays[ii]);
			return;
		}
		if ( verbose == 1 )
			printf( "Display 0x%x\n", (unsigned int)displays[ii]);

		if ( shouldShowAll == 1 )
			allModesForDisplay( displays[ii], verbose );

		if ( shouldFindClosest == 1 )
			findClosestModeForDisplay( displays[ii], verbose );

		if ( shouldFindExact == 1 )
			findExactModeForDisplay( displays[ii], verbose );

		if ( shouldFindHighest == 1 )
			highestModeForDisplay( displays[ii], &myMode, verbose );

		if ( shouldSetDisplay == 1 )
			setdisplay( displays[ii], verbose );

	}
	exit(0);
}
