#!/usr/bin/perl -w

################################################################################
# xhooks - radmind_xhooks_conf.pl
#
# Copyright (c) 2008 University of Utah Student Computing Labs.
# All Rights Reserved.
#
# Permission to use, copy, modify, and distribute this software and
# its documentation for any purpose and without fee is hereby granted,
# provided that the above copyright notice appears in all copies and
# that both that copyright notice and this permission notice appear
# in supporting documentation, and that the name of The University
# of Utah not be used in advertising or publicity pertaining to
# distribution of the software without specific, written prior
# permission. This software is supplied as is without expressed or
# implied warranties of any kind.
#
################################################################################
#
# Version 1.1
# - Added IP variables
#
#
# This script is part of the Xhooks package.
#
# http://www.macos.utah.edu/xhooks
#
# It requires the perlplist.pl script 
#	http://james-mac.scl.utah.edu/docroot/downloads/software/perlplist.pl
# You will need to change the 
#
# This script reads the radmind command files and saves settings to several 
# config/preferences files that are to be used by other apps or scripts.  It
# doesn't actually change any system config files.
#
# This script should run immediately after radmind (no other scripts that require
# config info should run)!
#
# To use this script, just make sure the command files are up-to-date, which is
# why I run this script right after lapply (after a reboot), and make sure you
# perlplist.pl and the location set below.  Then just run it.
#
# It also finds like that contain "## Begin uncomment and save to <PATH>".  
# Everything after this line is uncommented and saved to <PATH> up to the end of
# the command file or until the text "##? End uncomment and save" is found.
#
# -----------------------
#  SETTING VARIABLES
# -----------------------
#
# Command files
# -------------
#
# To create settings, put this in your command file
#
#	# SOME_VARIABLE = Some Value
#
# or
#
#	# SOME_VARIABLE_ARRAY += 1st value
#	# SOME_VARIABLE_ARRAY += 2nd value
#
# NOTE: The lines MUST be commented out in the command file or it will generate
# a radmind error on the server! (ktcheck wont even succeed on the client)
#
# NOTE 2: Don't put special characters in the variables names.  Otherwise, things
# will fail.  The values can have anything except special characters like quotes,
# $, @, \, etc.  I haven't added escaping so those characters will be interpreted
# to mean something other than the char, which could lead to all the scripts
# failing.
#
# As of version 1.0 (2008.01.30) this script creates a plist file, a perl file,
# and a bash file.
#
#
# IP Variables
# ------------
#
# An optional way of specifying which machines get which variable is by listing
# the IP then the variable.  This is so that one command file can contain multiple
# settings depending on the IP.  This is kinda like radmind's special file support.
# 
# To use, specify an IP, range, or wildcard in () followed by the variable name and value.
# 
# EXAMPLES:
# 
#	# (10.0.1.1) SOME_VARIABLE = Some Value
#
#	# (10.0.1.1) SOME_VARIABLE_ARRAY += 1st value
#	# (10.0.1.1) SOME_VARIABLE_ARRAY += 2nd value
#
# * wildcard is supported. So this line will apply to every computer (which is basically
# the same as leaving the IP part off):
#
#	# (*) SOME_VARIABLE_ARRAY += 2nd value
#
# * applies at any level.  These are all valid:
#
#	# (10.*) SOME_VARIABLE = Some Value
#	# (10.0.*) SOME_VARIABLE = Some Value
#	# (10.0.1.*) SOME_VARIABLE = Some Value
#
# Ranges are also supported.  So these are all valid:
#
#	# (10-20) SOME_VARIABLE = Some Value
#	# (10.0-127) SOME_VARIABLE = Some Value
#	# (10.0.128-255) SOME_VARIABLE = Some Value
#	# (10.0.1.254-255) SOME_VARIABLE = Some Value
#
# -----------------------
#  READING VARIABLES
# -----------------------
#
# Cocoa apps
# ----------
#
# The format of the plist file looks like this
#
#	<key>SOME_VARIABLE</key>
#	<string>Some Value</string>
#
# 	<key>SOME_VARIABLE_ARRAY</key>
#	<array>
#		<string>1st value</string>
#		<string>2nd value</string>
#	</array>
#
# A Cocoa app can read the settings using code like this:
#
#	NSUserDefaults *defaults = [NSUserDefaults standardUserDefaults];
#	[defaults addSuiteNamed:@"Xhooks/edu.utah.scl.radmind_xhooks_conf"];
#	[defaults registerDefaults:[NSDictionary dictionaryWithObjectsAndKeys:
#		@"SOME_VARIABLE", @"default value for app",
#		nil]];
#
#	someVariable = [defaults stringForKey:@"SOME_VARIABLE"];
#
# If you have defined SOME_VARIABLE in the machine's command file then someVariable
# will be equal to the value defined in the command file.  Otherwise it will be
# "default value for app".
#
#
# Perl scripts
# ------------
#
# The format of the Perl include looks like this
#
#	$SOME_VARIABLE = "Some Value";
#
#	@SOME_VARIABLE_ARRAY = (
#		"1st value",
#		"2nd value",
#	);
#
# A Perl script can use the settings by using code like this
#
#	if ( -f "/Library/Preferences/Xhooks/edu.utah.scl.radmind_xhooks_conf.perl" ) {
#		use lib "/Library/Preferences/Xhooks";
#		require "edu.utah.scl.radmind_xhooks_conf.perl";
#	}
#
#	if ( defined $SOME_VARIABLE ) {
#		# Do something with $SOME_VARIABLE
#	}
#
#   if ( defined @SOME_VARIABLE_ARRAY and $#SOME_VARIABLE_ARRAY > -1 ) {
#   	# Do something with @SOME_VARIABLE_ARRAY
#   }
#
# BASH scripts
# ------------
#
# The format of the BASH include looks like this
#
#	SOME_VARIABLE="Some Value"
#
#	SOME_VARIABLE_ARRAY=(
#		"1st value"
#		"2nd value"
#	)
#
# A BASH script can read the settings like so
#
#	if [ -f "/Library/Preferences/Xhooks/edu.utah.scl.radmind_xhooks_conf.bash" ]; then
#		. "/Library/Preferences/Xhooks/edu.utah.scl.radmind_xhooks_conf.bash"
#	fi
#
#	if [ "$SOME_VARIABLE" != "" ]; then
#		# Do something with $SOME_VARIABLE
#	fi
#
#
#

#################################################

use Data::Dumper;
use Foundation;
use lib "/Library/Xhooks/Modules/xhooks/lib";
require "perlplist.pl";

my $radmind_xhooks_plist = "/Library/Preferences/Xhooks/edu.utah.scl.radmind_xhooks_conf.plist";
my $radmind_xhooks_bash = "/Library/Preferences/Xhooks/edu.utah.scl.radmind_xhooks_conf.bash";
my $radmind_xhooks_perl = "/Library/Preferences/Xhooks/edu.utah.scl.radmind_xhooks_conf.perl";

if ( ! -e "/Library/Preferences/Xhooks" ) {
  mkdir "/Library/Preferences/Xhooks";
}

my $radmind_dir = "/var/radmind/client";
my $read_whole_command_file = 1;
my $verbose = 0;
my $my_ip;
my $network_is_setup = 0;

sub parse_command_file {
	my ( $radmind_dir, $command_file ) = @_;
	my @lines;
	my @found;
	if ( defined $command_file ) {
		$command_file = "$radmind_dir/$command_file";
	} else {
		$command_file = "$radmind_dir/command.K";		
	}
	if ( open COMMAND_FILE, "<", $command_file ) {
		@lines = <COMMAND_FILE>;
		close COMMAND_FILE;
	} else {
		die "Could not read $command_file: $!";
	}
	my $saving_to_file = 0;
	foreach my $line ( @lines ) {
		if ( $saving_to_file ) {
			if ( $line =~ /^##? End uncomment and save/ ) {
				$saving_to_file = 0;
				close SAVEFILE;
			} else {
				my $line2 = substr( $line, 1 );
				print SAVEFILE $line2;
			}
		} else {
			if ( $line =~ /^##? Begin uncomment and save to (.*)/ ) {
				$saving_to_file = 1;
				if ( ! open SAVEFILE, ">$1" ) {
					print STDERR "Could not write to $1: $!.\n";
					$saving_to_file = 0;
				} else {
					print "Creating file $1.\n" if $verbose;
				}
				
			} else {
				if ( $line =~ /^k *(.*)/ ) {
					@found = ( @found, parse_command_file( $radmind_dir, $1 ) );
				} else {
					chomp $line;
					push @found, $line;
				}
			}
		
		}
	}
	if ( $saving_to_file ) {
		close SAVEFILE;
	}
	return @found;
}

########################################################################
########################################################################
# Read settings
########################################################################
########################################################################

my @all_of_command_file;

# Read the command file
if ( $read_whole_command_file ) {
  @all_of_command_file = parse_command_file( $radmind_dir, "command.K" );
} else {
  my $mainConfig = "$radmind_dir/command.K";
  if ( open ( FILE, "<$mainConfig") ) {
    @all_of_command_file = <FILE>;
    close (FILE);
  } else {
    die "Error opening file $mainConfig: $!";
  }
}  

# Parse the lines.
foreach my $line (@all_of_command_file) {
  if ( $line =~ /^# (\w+)\s+=\s*(.*)/ ) {
    # KEY = value
    set_variable( $1, $2 );
  } elsif ( $line =~ /^# (\w+)\s+\+=\s*(.*)/ ) {
    # KEY += value
    add_to_array( $1, $2 );
  } elsif ( $line =~ /^# \(([\d\.\*-]*)\)\s+(\w+)\s+=\s*(.*)/ ) {
    # (IP) KEY = value
    my ( $compare_ip, $key, $value ) = ($1, $2, $3);
    if ( verify_network_is_setup() ) {
      if ( compare_ips( $compare_ip, $my_ip ) ) { 
        set_variable( $key, $value );
      }
    } else {
      print STDERR "I have a setting for with an IP but I can't get this machines IP: $line\n";
    }
  } elsif ( $line =~ /^# \(([\d\.\*-]*)\)\s+(\w+)\s+\+=\s*(.*)/ ) {
    # (IP) KEY += value
    my ( $compare_ip, $key, $value ) = ($1, $2, $3);
    if ( verify_network_is_setup() ) {
      if ( compare_ips( $compare_ip, $my_ip ) ) { 
        add_to_array( $key, $value );
      }
    } else {
      print STDERR "I have a setting for with an IP but I can't get this machines IP: $line\n";
    }
  }
}

sub set_variable {
  my ( $key, $value ) = @_;
  print "# $key = $value\n" if $verbose;
  if ( exists $keywords{$key} and $keywords{$key} ne '' ) {
    print STDERR "The key, $key, is defined twice.\n";
  }
  $keywords{$key} = $value;
}

sub add_to_array {
  my ( $key, $value ) = @_;
  print "# $key += $value\n" if $verbose;
  my $tempy = [];
  if ( exists $keywords{$key} ) {
    $tempy = $keywords{$key};
  }
  $keywords{$key} = [ @$tempy, $value ];
}

sub verify_network_is_setup {
  if ( ! $network_is_setup ) {
    use lib "/Library/Xhooks/Modules";
    require "xhooks/lib/common.pl";
    wait_for_network();
    $my_ip = get_ip();
    if ( $my_ip eq "" ) {
      print STDERR "I don't have an ip!\n";
    } else {
      $network_is_setup = 1;
    }
  }
  return $network_is_setup;
}

########################################################################
########################################################################
# Save Plist file
########################################################################
########################################################################

my $plist = cocoaDictFromPerlHash( %keywords );

print "Creating $radmind_xhooks_plist.\n" if $verbose;
printObject($plist) if $verbose;

saveDefaults( $plist, $radmind_xhooks_plist );
chmod 0640, $radmind_xhooks_plist;

########################################################################
########################################################################
# Save shell include file
########################################################################
########################################################################

print "Creating $radmind_xhooks_bash.\n" if $verbose;

if ( ! open SAVEFILE, ">$radmind_xhooks_bash" ) {
	print STDERR "Could not write to $radmind_xhooks_bash: $!.\n";
} else {
	while ( my ( $key, $value ) = each %keywords ) {
		if ( ref ( $value ) eq "ARRAY" ) {
			print "$key=(\n" if $verbose;
			print SAVEFILE "$key=(\n";
			foreach my $item ( @$value ) {
				if ( $item !~ /"/ ) {
					print "\t\"$item\"\n" if $verbose;
					print SAVEFILE "\t\"$item\"\n";
				} elsif ( $item !~ /\'/ ) {
					print "\t'$item'\n" if $verbose;
					print SAVEFILE "\t'$item'\n";
				} else {
					print STDERR "Something has quotes in it and it can't be written to bash config file.\n";
				}
			}
			print ")\n" if $verbose;
			print SAVEFILE ")\n";
		} else {
			if ( $value !~ /"/ ) {
				print "$key=\"$value\"\n" if $verbose;
				print SAVEFILE "$key=\"$value\"\n";
			} elsif ( $value !~ /'/ ) {
				print "$key='$value'\n" if $verbose;
				print SAVEFILE "$key='$value'\n";
			} else {
				print STDERR "Something has quotes in it and it can't be written to bash config file.\n";
			}
		}	
	}
	close SAVEFILE;
	chmod 0640, $radmind_xhooks_bash;
}

########################################################################
########################################################################
# Save perl include file
########################################################################
########################################################################

print "Creating $radmind_xhooks_perl.\n" if $verbose;

if ( ! open SAVEFILE, ">$radmind_xhooks_perl" ) {
	print STDERR "Could not write to $radmind_xhooks_perl: $!.\n";
} else {
	while ( my ( $key, $value ) = each %keywords ) {
		if ( ref ( $value ) eq "ARRAY" ) {
			print "\@$key = (\n" if $verbose;
			print SAVEFILE "\@$key = (\n";
			foreach my $item ( @$value ) {
				if ( $item !~ /'/ ) {
					print "\t'$item',\n" if $verbose;
					print SAVEFILE "\t'$item',\n";
				} elsif ( $item !~ /\"/ ) {
					print "\t\"$item\",\n" if $verbose;
					print SAVEFILE "\t\"$item\",\n";
				} else {
					print STDERR "Something has quotes in it and it can't be written to perl config file.\n";
				}
			}
			print ");\n" if $verbose;
			print SAVEFILE ");\n";
		} else {
			if ( $value !~ /'/ ) {
				print "\$$key = '$value';\n" if $verbose;
				print SAVEFILE "\$$key = '$value';\n";
			} elsif ( $value !~ /"/ ) {
				print "\$$key = \"$value\";\n" if $verbose;
				print SAVEFILE "\$$key = \"$value\";\n";
			} else {
				print STDERR "Something has quotes in it and it can't be written to perl config file.\n";
			}
		}	
	}
	close SAVEFILE;
	chmod 0640, $radmind_xhooks_perl;
}

exit 0;
